from __future__ import annotations

from typing import TYPE_CHECKING

import dagster as dg

DAGSTER_RAY_OBJECT_MAP_NAME = "DagsterRayObjectMap"
DAGSTER_RAY_NAMESPACE = "dagster-ray"

# we need to create a global Ray actor which will store all the refs to all objcets

if TYPE_CHECKING:
    import ray


class RayObjectMap:
    # TODO: implement some eventual cleanup mechanism
    # idea: save creation timestamp and periodically check for old refs
    # or add some integration with the RunLauncher/Executor
    def __init__(self):
        self._object_map: dict[str, ray.ObjectRef] = {}

    def set(self, key: str, ref: ray.ObjectRef):
        self._object_map[key] = ref

    def get(self, key: str) -> ray.ObjectRef | None:
        return self._object_map.get(key)

    def delete(self, key: str):
        if key in self._object_map:
            del self._object_map[key]

    def keys(self):
        return self._object_map.keys()

    def ping(self):
        return "pong"

    @staticmethod
    def get_or_create():
        import ray

        actor = (
            ray.remote(RayObjectMap)
            .options(  # type: ignore
                name=DAGSTER_RAY_OBJECT_MAP_NAME,
                namespace=DAGSTER_RAY_NAMESPACE,
                get_if_exists=True,
                lifetime="detached",
                # max_restarts=-1,
                max_concurrency=1000,  # TODO: make this configurable,
                runtime_env={"RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING": "1"},
            )
            .remote()
        )

        # make sure the actor is created
        ray.get(actor.ping.remote())  # type: ignore

        return actor


class RayIOManager(dg.ConfigurableIOManager):
    """IO Manager that stores intermediate values in Ray's object store.

    The RayIOManager allows storing and retrieving intermediate values in Ray's distributed
    object store, making it ideal for use with RayRunLauncher and ray_executor. It works by
    storing Dagster step keys in a global Ray actor that maintains a mapping between step
    keys and Ray ObjectRefs.

    Args:
        address: Ray cluster address. If provided, will initialize Ray connection.
                If None, assumes Ray is already initialized.

    Example:
        Basic usage
        ```python
        import dagster as dg
        from dagster_ray import RayIOManager

        @dg.asset(io_manager_key="ray_io_manager")
        def upstream() -> int:
            return 42

        @dg.asset
        def downstream(upstream: int):
            return upstream * 2

        definitions = dg.Definitions(
            assets=[upstream, downstream],
            resources={"ray_io_manager": RayIOManager()}
        )
        ```

    Example:
        With Ray cluster address
        ```python
        ray_io_manager = RayIOManager(address="ray://head-node:10001")
        ```

    Info:
        - Works with any pickable Python objects
        - Supports partitioned assets and partition mappings
        - Uses Ray's automatic object movement for fault tolerance
        - Objects are stored with the Ray actor as owner for lifecycle management
    """

    address: str | None = None

    def handle_output(self, context: dg.OutputContext, obj):
        import ray

        if self.address:  # TODO: should this really be done here?
            ray.init(self.address, ignore_reinit_error=True)

        object_map = RayObjectMap.get_or_create()

        storage_key = self._get_single_key(context)

        # TODO: understand if Ray will automatically move the object from dying nodes
        # what if not?

        ref = ray.put(obj, _owner=object_map)

        object_map.set.remote(storage_key, ref)  # type: ignore

        context.log.debug(f"[RayIOManager] Stored object with key {storage_key} as {ref}")

    def load_input(self, context: dg.InputContext):
        import ray

        if self.address:  # TODO: should this really be done here?
            ray.init(self.address, ignore_reinit_error=True)

        object_map = RayObjectMap.get_or_create()

        if context.has_asset_partitions and len(context.asset_partition_keys) > 1:
            # load multiple partitions as once
            # first, get the refs

            storage_keys = self._get_multiple_keys(context)
            refs = [object_map.get.remote(key) for key in storage_keys.values()]  # type: ignore
            values = ray.get(refs)
            return {partition_key: value for partition_key, value in zip(storage_keys.keys(), values)}

        else:
            storage_key = self._get_single_key(context)

        context.log.debug(f"[RayIOManager] Loading object with key {storage_key}")

        ref = object_map.get.remote(storage_key)  # type: ignore

        assert ref is not None, f"[RayIOManager] Object with key {storage_key} not found in RayObjectMap"

        return ray.get(ref)

    def _get_single_key(self, context: dg.InputContext | dg.OutputContext) -> str:
        identifier = context.get_identifier() if not context.has_asset_key else context.get_asset_identifier()
        return "/".join(identifier)

    def _get_multiple_keys(self, context: dg.InputContext) -> dict[str, str]:
        if context.has_asset_key:
            asset_path = list(context.asset_key.path)

            return {
                partition_key: "/".join(asset_path + [partition_key]) for partition_key in context.asset_partition_keys
            }
        else:
            raise RuntimeError(
                "[RayIOManager] This method can only be called with an InputContext that has multiple partitions"
            )



import socket
import sys
import time
from pathlib import Path
from typing import Any, Literal, cast

import dagster as dg
import pytest
import ray  # noqa: TID253
from packaging.version import Version
from pytest_kubernetes.providers import AClusterManager

from dagster_ray import RayResource
from dagster_ray._base.resources import Lifecycle
from dagster_ray.kuberay import (
    KubeRayCluster,
    KubeRayInteractiveJob,
)
from dagster_ray.kuberay.client.rayjob.client import RayJobClient
from dagster_ray.kuberay.configs import RayClusterSpec
from dagster_ray.kuberay.resources.rayjob import InteractiveRayJobConfig, InteractiveRayJobSpec
from tests.kuberay.conftest import KUBERNETES_CONTEXT, RAYJOB_TIMEOUT
from tests.kuberay.utils import NAMESPACE, get_random_free_port

MIN_KUBERAY_VERSION = "1.3.0"


@pytest.fixture(scope="session")
def rayjob_client(k8s_with_kuberay: AClusterManager) -> RayJobClient:
    return RayJobClient(kube_config=str(k8s_with_kuberay.kubeconfig), kube_context=KUBERNETES_CONTEXT)


def test_instantiate_defaults():
    _ = KubeRayInteractiveJob()


def test_no_lifecycle(dagster_instance: dg.DagsterInstance, rayjob_client: RayJobClient):
    interactive_rayjob = KubeRayInteractiveJob(
        client=rayjob_client, lifecycle=Lifecycle(create=False, wait=False, connect=False)
    )

    @dg.asset
    def my_asset(interactive_rayjob: KubeRayInteractiveJob) -> None:
        return

    dg.materialize(assets=[my_asset], resources={"interactive_rayjob": interactive_rayjob}, instance=dagster_instance)


@pytest.mark.parametrize("cleanup", ["always", "never"])
@pytest.mark.parametrize("wait", [True, False])
def test_cleanup(
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    k8s_with_kuberay: AClusterManager,
    kuberay_version: str,
    dagster_instance: dg.DagsterInstance,
    rayjob_client: RayJobClient,
    dagster_ray_image: str,
    wait: bool,
    cleanup: Literal["always", "never", "on_exception"],
):
    if Version(kuberay_version) < Version(MIN_KUBERAY_VERSION):
        pytest.skip(f"KubeRay {MIN_KUBERAY_VERSION} is required to use interactive mode with RayJob")

    interactive_rayjob = KubeRayInteractiveJob(
        image=dagster_ray_image,
        client=rayjob_client,
        redis_port=get_random_free_port(),
        ray_job=InteractiveRayJobConfig(
            metadata={"namespace": NAMESPACE},
            spec=InteractiveRayJobSpec(
                ray_cluster_spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
            ),
        ),
        lifecycle=Lifecycle(create=True, wait=wait, connect=False, cleanup=cleanup),
        timeout=RAYJOB_TIMEOUT,
    )

    @dg.asset
    def my_asset(interactive_rayjob: KubeRayInteractiveJob) -> None:
        return

    res = dg.materialize(
        assets=[my_asset],
        resources={"interactive_rayjob": interactive_rayjob},
        instance=dagster_instance,
        raise_on_error=False,
    )

    # Wait for cleanup to complete with retry mechanism
    max_wait_time = 30  # seconds
    wait_interval = 1  # seconds
    start_time = time.time()

    while time.time() - start_time < max_wait_time:
        jobs = rayjob_client.list(
            namespace=interactive_rayjob.namespace, label_selector=f"dagster/run-id={res.run_id}"
        )["items"]

        # Filter out jobs that are being deleted (have deletionTimestamp)
        active_jobs = [job for job in jobs if job.get("metadata", {}).get("deletionTimestamp") is None]

        if cleanup == "always" and len(active_jobs) == 0:
            break
        elif cleanup == "never" and len(active_jobs) == 1:
            break

        time.sleep(wait_interval)

    # Final check
    jobs = rayjob_client.list(namespace=interactive_rayjob.namespace, label_selector=f"dagster/run-id={res.run_id}")[
        "items"
    ]
    active_jobs = [job for job in jobs if job.get("metadata", {}).get("deletionTimestamp") is None]

    if cleanup == "always":
        assert len(active_jobs) == 0, (
            f"Expected 0 active jobs after cleanup, but found {len(active_jobs)} active jobs out of {len(jobs)} total jobs"
        )
    elif cleanup == "never":
        assert len(active_jobs) == 1, (
            f"Expected 1 active job when cleanup is disabled, but found {len(active_jobs)} active jobs out of {len(jobs)} total jobs"
        )


@pytest.mark.parametrize("cleanup", ["always", "on_exception"])
def test_cleanup_on_interrupt(
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    k8s_with_kuberay: AClusterManager,
    kuberay_version: str,
    dagster_instance: dg.DagsterInstance,
    rayjob_client: RayJobClient,
    cleanup: str,
):
    import json
    import signal
    import subprocess

    if Version(kuberay_version) < Version(MIN_KUBERAY_VERSION):
        pytest.skip(f"KubeRay {MIN_KUBERAY_VERSION} is required to use interactive mode with RayJob")

    # Path to the separate test script
    script_path = str(Path(__file__).parents[1] / "scripts/launch_dagster_run_with_kuberay_interactive_job.py")

    assert Path(script_path).exists()

    # Build command line arguments
    cmd_args = [
        sys.executable,
        script_path,
        "--config-file",
        str(k8s_with_kuberay.kubeconfig),
        "--context",
        KUBERNETES_CONTEXT,
        "--image",
        "invalid-image:nonexistent",  # this will cause the RayJob to hang, we need it to test the interrupt cleanup functionality
        "--redis-port",
        str(get_random_free_port()),
        "--namespace",
        NAMESPACE,
        "--head-group-spec",
        json.dumps(head_group_spec),
        "--worker-group-specs",
        json.dumps(worker_group_specs),
        "--cleanup",
        cleanup,
    ]

    # Start the subprocess with stdout capture
    process = subprocess.Popen(cmd_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

    # Wait a bit to let the RayJob get created
    time.sleep(10)

    # Interrupt the process
    process.send_signal(signal.SIGTERM)

    # Wait for process to terminate and capture output
    try:
        stdout, stderr = process.communicate(timeout=30)
    except subprocess.TimeoutExpired:
        process.kill()
        stdout, stderr = process.communicate()

    # Extract run ID from stdout (last line should be the run ID)
    assert stdout, "Expected to get stdout from the subprocess"
    lines = stdout.strip().split("\n")
    run_id = None
    for line in reversed(lines):
        if line.strip() and "-" in line.strip():  # Run IDs have dashes
            run_id = line.strip()
            break

    assert run_id, f"Expected to get a run ID from subprocess output, got: {stdout}"

    # Wait for cleanup to complete with retry mechanism
    max_wait_time = 30  # seconds
    wait_interval = 1  # seconds
    start_time = time.time()

    while time.time() - start_time < max_wait_time:
        jobs = rayjob_client.list(namespace=NAMESPACE, label_selector=f"dagster/run-id={run_id}")["items"]

        # Filter out jobs that are being deleted (have deletionTimestamp)
        active_jobs = [job for job in jobs if job.get("metadata", {}).get("deletionTimestamp") is None]

        if len(active_jobs) == 0:
            break

        time.sleep(wait_interval)

    # Final verification that RayJob has been cleaned up for this specific run
    final_jobs = rayjob_client.list(namespace=NAMESPACE, label_selector=f"dagster/run-id={run_id}")["items"]
    active_final_jobs = [job for job in final_jobs if job.get("metadata", {}).get("deletionTimestamp") is None]

    # The cleanup should have removed the RayJob created by this test
    assert len(active_final_jobs) == 0, (
        f"RayJob with run-id {run_id} should have been cleaned up, but found {len(active_final_jobs)} active jobs out of {len(final_jobs)} total jobs"
    )

    # Ensure no leftover jobs in case test failed
    try:
        jobs = rayjob_client.list(namespace=NAMESPACE)["items"]
        for job in jobs:
            rayjob_client.delete(job["metadata"]["name"], namespace=NAMESPACE)
    except Exception:
        pass  # Best effort cleanup


@pytest.fixture(scope="session")
def interactive_rayjob_resource(
    rayjob_client: RayJobClient,
    dagster_ray_image: str,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    kuberay_version: str,
) -> KubeRayInteractiveJob:
    if Version(kuberay_version) < Version(MIN_KUBERAY_VERSION):
        pytest.skip(f"KubeRay {MIN_KUBERAY_VERSION} is required to use interactive mode with RayJob")

    return KubeRayInteractiveJob(
        image=dagster_ray_image,
        client=rayjob_client,
        ray_job=InteractiveRayJobConfig(
            metadata={"namespace": NAMESPACE},
            spec=InteractiveRayJobSpec(
                ray_cluster_spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
            ),
        ),
        redis_port=get_random_free_port(),
        lifecycle=Lifecycle(connect=False),
        timeout=RAYJOB_TIMEOUT,
    )


@ray.remote
def get_hostname():
    return socket.gethostname()


def ensure_interactive_rayjob_correctness(
    rayjob: KubeRayInteractiveJob,
    k8s_with_kuberay: AClusterManager,
    context: dg.AssetExecutionContext,
):
    with k8s_with_kuberay.port_forwarding(
        target=f"svc/{rayjob.cluster_name}-head-svc",
        source_port=cast(int, rayjob.redis_port),
        target_port=10001,
        namespace=rayjob.namespace,
    ):
        assert rayjob.client.get_status(rayjob.name, rayjob.namespace)["jobDeploymentStatus"] == "Waiting"  # pyright: ignore[reportTypedDictNotRequiredAccess]

        # now we can access the head node
        # hack the _host attribute to point to the port-forwarded address
        rayjob._host = "127.0.0.1"
        rayjob.connect(context)  # normally this would happen automatically during resource setup
        assert rayjob.context is not None

        time.sleep(1)

        job_status = rayjob.client.get_status(rayjob.name, rayjob.namespace).get("jobStatus")
        assert job_status == "RUNNING", job_status

        # make sure a @remote function runs inside the cluster
        # not in localhost
        assert rayjob.cluster_name in ray.get(get_hostname.remote())

        rayjob_description = rayjob.client.get(rayjob.name, namespace=rayjob.namespace)
        assert rayjob_description["metadata"]["labels"]["dagster/run-id"] == context.run_id


def test_interactive_rayjob(
    interactive_rayjob_resource: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
):
    @dg.asset
    # testing RayResource type annotation too!
    def my_asset(context: dg.AssetExecutionContext, interactive_rayjob: RayResource) -> None:
        # port-forward to the head node
        # because it's not possible to access it otherwise

        assert isinstance(interactive_rayjob, KubeRayInteractiveJob)

        ensure_interactive_rayjob_correctness(
            interactive_rayjob,
            k8s_with_kuberay,
            context,
        )

    dg.materialize_to_memory(
        [my_asset],
        resources={"interactive_rayjob": interactive_rayjob_resource},
    )



import logging
from collections.abc import Iterator

import dagster as dg
import pytest
from _pytest.tmpdir import TempPathFactory

logging.getLogger("alembic").setLevel(logging.WARNING)


@pytest.fixture
def dagster_instance(tmp_path_factory: TempPathFactory) -> dg.DagsterInstance:
    return dg.DagsterInstance.ephemeral(tempdir=str(tmp_path_factory.mktemp("dagster_home")))


@pytest.fixture
def local_ray_address() -> Iterator[str]:
    import ray

    context = ray.init(
        ignore_reinit_error=True, runtime_env={"env_vars": {"RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING": "1"}}
    )

    yield "auto"

    context.disconnect()



from dagster_ray.kuberay.configs import RayClusterConfig, RayClusterSpec, RayJobConfig, RayJobSpec
from dagster_ray.kuberay.jobs import cleanup_kuberay_clusters, delete_kuberay_clusters
from dagster_ray.kuberay.ops import cleanup_kuberay_clusters_op, delete_kuberay_clusters_op
from dagster_ray.kuberay.pipes import PipesKubeRayJobClient
from dagster_ray.kuberay.resources import (
    KubeRayCluster,
    KubeRayClusterClientResource,
    KubeRayInteractiveJob,
    KubeRayJobClientResource,
)
from dagster_ray.kuberay.schedules import cleanup_kuberay_clusters_daily

__all__ = [
    "KubeRayCluster",
    "RayClusterConfig",
    "KubeRayClusterClientResource",
    "PipesKubeRayJobClient",
    "cleanup_kuberay_clusters",
    "delete_kuberay_clusters",
    "cleanup_kuberay_clusters_op",
    "delete_kuberay_clusters_op",
    "cleanup_kuberay_clusters_daily",
    "RayClusterSpec",
    "RayJobConfig",
    "RayJobSpec",
    "KubeRayInteractiveJob",
    "KubeRayJobClientResource",
]



#!/usr/bin/env python3
"""
Script for testing process interrupt cleanup behavior.
This script creates a KubeRayInteractiveJob that will hang due to invalid image,
allowing the parent test to interrupt it and verify cleanup behavior.
"""

import argparse
import json
import sys

import dagster as dg

from dagster_ray._base.resources import Lifecycle
from dagster_ray.kuberay import KubeRayInteractiveJob
from dagster_ray.kuberay.client.rayjob.client import RayJobClient
from dagster_ray.kuberay.configs import RayClusterSpec
from dagster_ray.kuberay.resources.rayjob import InteractiveRayJobConfig, InteractiveRayJobSpec


def parse_args():
    parser = argparse.ArgumentParser(description="Test script for process interrupt cleanup")
    parser.add_argument("--config-file", required=True, help="Path to kubeconfig file")
    parser.add_argument("--context", required=True, help="Kubernetes context name")
    parser.add_argument("--image", required=True, help="Docker image name")
    parser.add_argument("--cleanup", required=True, help="Cleanup behavior")
    parser.add_argument("--redis-port", type=int, required=True, help="Redis port number")
    parser.add_argument("--namespace", required=True, help="Kubernetes namespace")
    parser.add_argument("--head-group-spec", required=True, help="Head group spec as JSON string")
    parser.add_argument("--worker-group-specs", required=True, help="Worker group specs as JSON string")
    return parser.parse_args()


def main():
    args = parse_args()

    # Parse JSON arguments
    try:
        head_group_spec = json.loads(args.head_group_spec)
        worker_group_specs = json.loads(args.worker_group_specs)
    except json.JSONDecodeError as e:
        print(f"Error parsing JSON arguments: {e}", file=sys.stderr)
        sys.exit(1)

    # Initialize the client and resource
    client = RayJobClient(kube_config=args.config_file, kube_context=args.context)

    interactive_rayjob = KubeRayInteractiveJob(
        image=args.image,  # Invalid image that will cause hanging
        client=client,
        redis_port=args.redis_port,
        ray_job=InteractiveRayJobConfig(
            metadata={"namespace": args.namespace},
            spec=InteractiveRayJobSpec(
                ray_cluster_spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
            ),
        ),
        lifecycle=Lifecycle(create=True, wait=True, connect=False, cleanup=args.cleanup),
        timeout=300,  # Large timeout
    )

    @dg.asset
    def hanging_asset(interactive_rayjob: KubeRayInteractiveJob) -> None:
        return

    res = dg.materialize(
        assets=[hanging_asset],
        resources={"interactive_rayjob": interactive_rayjob},
        raise_on_error=False,
    )

    # we print the run id so the caller can use it to filter by RayJob `dagster/run-id` label in kubernetes
    print(res.run_id)


if __name__ == "__main__":
    main()



import sys
from typing import Optional

import dagster as dg
from pydantic import Field, PrivateAttr
from typing_extensions import override

from dagster_ray.kuberay.client import RayClusterClient
from dagster_ray.kuberay.configs import RayClusterConfig
from dagster_ray.kuberay.resources.base import BaseKubeRayResourceConfig
from dagster_ray.kuberay.utils import normalize_k8s_label_values
from dagster_ray.types import AnyDagsterContext

if sys.version_info >= (3, 11):
    pass
else:
    pass

from ray._private.worker import BaseContext as RayBaseContext  # noqa

from dagster_ray._base.resources import BaseRayResource, Lifecycle
from dagster_ray.kuberay.client.base import load_kubeconfig


class KubeRayClusterClientResource(dg.ConfigurableResource[RayClusterClient]):
    """This configurable resource provides a `dagster_ray.kuberay.client.RayClusterClient`."""

    kube_context: Optional[str] = None
    kube_config: Optional[str] = None

    def create_resource(self, context: dg.InitResourceContext) -> RayClusterClient:
        load_kubeconfig(context=self.kube_context, config_file=self.kube_config)

        return RayClusterClient(kube_context=self.kube_context, kube_config=self.kube_config)


class KubeRayCluster(BaseKubeRayResourceConfig, BaseRayResource):
    """
    Provides a `RayCluster` for Dagster steps.

    It is advised to use `dagster_ray.kuberay.KubeRayInteractiveJob` with KubeRay >= 1.3.0 instead.

    Info:
        Image defaults to `dagster/image` run tag.

    Tip:
        Make sure `ray[full]` is available in the image.
    """

    lifecycle: Lifecycle = Field(
        default_factory=lambda: Lifecycle(cleanup="always"), description="Actions to perform during resource setup."
    )

    ray_cluster: RayClusterConfig = Field(
        default_factory=RayClusterConfig, description="Kubernetes `RayCluster` CR configuration."
    )
    client: dg.ResourceDependency[RayClusterClient] = Field(  # pyright: ignore[reportAssignmentType]
        default_factory=KubeRayClusterClientResource, description="Kubernetes `RayCluster` client"
    )
    log_cluster_conditions: bool = Field(
        default=True,
        description="Whether to log RayCluster conditions while waiting for the RayCluster to become ready. For more information, see https://docs.ray.io/en/latest/cluster/kubernetes/user-guides/observability.html#raycluster-status-conditions.",
    )

    _name: str = PrivateAttr()
    _host: str = PrivateAttr()

    @property
    def host(self) -> str:
        if not hasattr(self, "_host"):
            raise ValueError(f"{self.__class__.__name__} not initialized")
        return self._host

    @property
    def name(self) -> str:
        if not hasattr(self, "_name") and self.ray_cluster.metadata.get("name") is None:
            raise ValueError(f"{self.__class__.__name__} not initialized")
        elif (name := self.ray_cluster.metadata.get("name")) is not None:
            return name
        else:
            return self._name

    @property
    def namespace(self) -> str:
        return self.ray_cluster.namespace

    @property
    @override
    def display_name(self) -> str:
        return (
            f"RayCluster {self.namespace}/{self.name}" if self.created else f"RayCluster in namespace {self.namespace}"
        )

    def get_dagster_tags(self, context: AnyDagsterContext) -> dict[str, str]:
        tags = super().get_dagster_tags(context=context)
        tags.update({"dagster/deployment": self.deployment_name})
        return tags

    @override
    def create(self, context: AnyDagsterContext):
        assert context.log is not None
        assert context.dagster_run is not None

        self._name = self.ray_cluster.metadata.get("name") or self._get_step_name(context)

        # just a safety measure, no need to recreate the cluster for step retries or smth
        if not self.client.get(
            name=self.name,
            namespace=self.namespace,
        ):
            k8s_manifest = self.ray_cluster.to_k8s(
                context,
                image=(self.image or context.dagster_run.tags.get("dagster/image")),
                labels=normalize_k8s_label_values(self.get_dagster_tags(context)),
                env_vars=self.get_env_vars_to_inject(),
            )

            k8s_manifest["metadata"]["name"] = self.name

            resource = self.client.create(body=k8s_manifest, namespace=self.namespace)
            if not resource:
                raise RuntimeError(f"Couldn't create {self.display_name}")

    @override
    def wait(self, context: AnyDagsterContext):
        assert context.log is not None
        assert context.dagster_run is not None

        self.client.wait_until_ready(
            self.name,
            namespace=self.namespace,
            timeout=self.timeout,
            poll_interval=self.poll_interval,
            log_cluster_conditions=self.log_cluster_conditions,
        )

        self._host = self.client.get_status(
            name=self.name, namespace=self.namespace, timeout=self.timeout, poll_interval=self.poll_interval
        )[  # pyright: ignore
            "head"
        ]["serviceIP"]

        msg = f"RayCluster {self.namespace}/{self.name} is ready! Connection command:\n"
        msg += f"kubectl -n {self.namespace} port-forward svc/{self.name}-head-svc 8265:8265 6379:6379 10001:10001"

        context.log.info(msg)

    @override
    def delete(self, context: AnyDagsterContext):
        self.client.delete(self.name, namespace=self.namespace)



import socket
from typing import Any, cast

import dagster as dg
import pytest
import ray  # noqa: TID253
from pytest_kubernetes.providers import AClusterManager

from dagster_ray import Lifecycle, RayResource
from dagster_ray.kuberay import KubeRayCluster, KubeRayClusterClientResource, RayClusterConfig, cleanup_kuberay_clusters
from dagster_ray.kuberay.client import RayClusterClient
from dagster_ray.kuberay.configs import RayClusterSpec
from dagster_ray.kuberay.ops import CleanupKuberayClustersConfig
from tests.kuberay.conftest import KUBERNETES_CONTEXT
from tests.kuberay.utils import NAMESPACE, get_random_free_port


@pytest.fixture(scope="session")
def raycluster_client(k8s_with_kuberay: AClusterManager):
    return KubeRayClusterClientResource(kube_config=str(k8s_with_kuberay.kubeconfig), kube_context=KUBERNETES_CONTEXT)


@pytest.fixture(scope="session")
def ray_cluster_resource(
    k8s_with_kuberay: AClusterManager,
    raycluster_client: KubeRayClusterClientResource,
    dagster_ray_image: str,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
) -> KubeRayCluster:
    redis_port = get_random_free_port()

    return KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(connect=False),
        client=raycluster_client,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=redis_port,
    )


@pytest.fixture(scope="session")
def ray_cluster_resource_skip_cleanup(
    k8s_with_kuberay: AClusterManager,
    dagster_ray_image: str,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    raycluster_client: KubeRayClusterClientResource,
) -> KubeRayCluster:
    redis_port = get_random_free_port()

    return KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(
            connect=False,
            cleanup="never",
        ),
        client=raycluster_client,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=redis_port,
    )


@pytest.fixture(scope="session")
def ray_cluster_resource_skip_create(
    k8s_with_kuberay: AClusterManager,
    dagster_ray_image: str,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    raycluster_client: KubeRayClusterClientResource,
) -> KubeRayCluster:
    redis_port = get_random_free_port()

    return KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(create=False),
        client=raycluster_client,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=redis_port,
    )


@pytest.fixture(scope="session")
def ray_cluster_resource_skip_wait(
    k8s_with_kuberay: AClusterManager,
    dagster_ray_image: str,
    head_group_spec: dict[str, Any],
    worker_group_specs: list[dict[str, Any]],
    raycluster_client: KubeRayClusterClientResource,
) -> KubeRayCluster:
    redis_port = get_random_free_port()

    return KubeRayCluster(
        image=dagster_ray_image,
        # have have to first run port-forwarding with minikube
        # we can only init ray after that
        lifecycle=Lifecycle(wait=False),
        client=raycluster_client,
        ray_cluster=RayClusterConfig(
            metadata={"namespace": NAMESPACE},
            spec=RayClusterSpec(head_group_spec=head_group_spec, worker_group_specs=worker_group_specs),
        ),
        redis_port=redis_port,
    )


@ray.remote
def get_hostname():
    return socket.gethostname()


def ensure_kuberay_cluster_correctness(
    ray_cluster: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
    context: dg.AssetExecutionContext,
):
    with k8s_with_kuberay.port_forwarding(
        target=f"svc/{ray_cluster.name}-head-svc",
        source_port=cast(int, ray_cluster.redis_port),
        target_port=10001,
        namespace=ray_cluster.namespace,
    ):
        # now we can access the head node
        # hack the _host attribute to point to the port-forwarded address
        ray_cluster._host = "127.0.0.1"
        ray_cluster.connect(context)  # normally this would happen automatically during resource setup
        assert ray_cluster.context is not None

        # make sure a @remote function runs inside the cluster
        # not in localhost
        assert ray_cluster.name in ray.get(get_hostname.remote())

        ray_cluster_description = ray_cluster.client.get(ray_cluster.name, namespace=ray_cluster.namespace)
        assert ray_cluster_description["metadata"]["labels"]["dagster/run-id"] == context.run_id


def test_kuberay_cluster_resource(
    ray_cluster_resource: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
):
    @dg.asset
    # testing RayResource type annotation too!
    def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource) -> None:
        # port-forward to the head node
        # because it's not possible to access it otherwise

        assert isinstance(ray_cluster, KubeRayCluster)

        ensure_kuberay_cluster_correctness(
            ray_cluster,
            k8s_with_kuberay,
            context,
        )

    result = dg.materialize_to_memory(
        [my_asset],
        resources={"ray_cluster": ray_cluster_resource},
    )

    kuberay_client = RayClusterClient(kube_config=str(k8s_with_kuberay.kubeconfig))

    # make sure the RayCluster is cleaned up

    assert (
        len(
            kuberay_client.list(
                namespace=ray_cluster_resource.namespace, label_selector=f"dagster/run-id={result.run_id}"
            )["items"]
        )
        == 0
    )


def test_kuberay_cluster_resource_skip_create(
    ray_cluster_resource_skip_create: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
):
    @dg.asset
    def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource) -> None:
        assert isinstance(ray_cluster, KubeRayCluster)

        # call create and wait manually
        ray_cluster.create(context)
        ray_cluster.wait(context)

        ensure_kuberay_cluster_correctness(
            ray_cluster,
            k8s_with_kuberay,
            context,
        )

    dg.materialize_to_memory(
        [my_asset],
        resources={"ray_cluster": ray_cluster_resource_skip_create},
    )


def test_kuberay_cluster_resource_skip_wait(
    ray_cluster_resource_skip_wait: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
):
    @dg.asset
    def my_asset(context: dg.AssetExecutionContext, ray_cluster: RayResource) -> None:
        assert isinstance(ray_cluster, KubeRayCluster)

        # call wait manually
        ray_cluster.wait(context)

        ensure_kuberay_cluster_correctness(
            ray_cluster,
            k8s_with_kuberay,
            context,
        )

    dg.materialize_to_memory(
        [my_asset],
        resources={"ray_cluster": ray_cluster_resource_skip_wait},
    )


def test_kuberay_cleanup_job(
    ray_cluster_resource_skip_cleanup: KubeRayCluster,
    k8s_with_kuberay: AClusterManager,
):
    @dg.asset
    def my_asset(ray_cluster: RayResource) -> None:
        assert isinstance(ray_cluster, KubeRayCluster)

    result = dg.materialize_to_memory(
        [my_asset],
        resources={"ray_cluster": ray_cluster_resource_skip_cleanup},
    )

    raycluster_client = RayClusterClient(kube_config=str(k8s_with_kuberay.kubeconfig), kube_context=KUBERNETES_CONTEXT)

    assert (
        len(
            raycluster_client.list(
                namespace=ray_cluster_resource_skip_cleanup.namespace,
                label_selector=f"dagster/run-id={result.run_id}",
            )["items"]
        )
        > 0
    )

    cleanup_kuberay_clusters.execute_in_process(
        resources={
            "kuberay_client": KubeRayClusterClientResource(kube_config=str(k8s_with_kuberay.kubeconfig)),
        },
        run_config=dg.RunConfig(
            ops={
                "cleanup_kuberay_clusters": CleanupKuberayClustersConfig(
                    namespace=ray_cluster_resource_skip_cleanup.namespace,
                )
            }
        ),
    )

    assert not raycluster_client.list(
        namespace=ray_cluster_resource_skip_cleanup.namespace, label_selector=f"dagster/run-id={result.run_id}"
    )["items"]


def test_ray_cluster_builder_debug():
    kuberay_cluster = KubeRayCluster(enable_debug_post_mortem=True, image="test")
    kuberay_cluster._cluster_name = "test-cluster"
    context = dg.build_init_resource_context()

    ray_cluster_config = kuberay_cluster.ray_cluster.to_k8s(
        context, env_vars=kuberay_cluster.get_env_vars_to_inject(), labels={"foo": "bar"}, image="test"
    )
    assert ray_cluster_config["metadata"]["labels"]["foo"] == "bar"
    for group_spec in [ray_cluster_config["spec"]["headGroupSpec"], *ray_cluster_config["spec"]["workerGroupSpecs"]]:
        for container in group_spec["template"]["spec"]["containers"]:
            assert {"name": "RAY_DEBUG_POST_MORTEM", "value": "1"} in container["env"], container

    kuberay_cluster = KubeRayCluster(enable_tracing=True)
    kuberay_cluster._cluster_name = "test-cluster"
    ray_cluster_config = kuberay_cluster.ray_cluster.to_k8s(
        context, env_vars=kuberay_cluster.get_env_vars_to_inject(), image="test"
    )
    for group_spec in [ray_cluster_config["spec"]["headGroupSpec"], *ray_cluster_config["spec"]["workerGroupSpecs"]]:
        for container in group_spec["template"]["spec"]["containers"]:
            assert {"name": "RAY_PROFILING", "value": "1"} in container["env"], container

    kuberay_cluster = KubeRayCluster(enable_actor_task_logging=True)
    kuberay_cluster._cluster_name = "test-cluster"
    ray_cluster_config = kuberay_cluster.ray_cluster.to_k8s(
        context, env_vars=kuberay_cluster.get_env_vars_to_inject(), image="test"
    )
    kuberay_cluster._cluster_name = "test-cluster"
    for group_spec in [ray_cluster_config["spec"]["headGroupSpec"], *ray_cluster_config["spec"]["workerGroupSpecs"]]:
        for container in group_spec["template"]["spec"]["containers"]:
            assert {"name": "RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING", "value": "1"} in container["env"], container



from dagster_ray.kuberay.configs import RayClusterConfig, RayClusterSpec, RayJobConfig, RayJobSpec
from dagster_ray.kuberay.resources.raycluster import (
    KubeRayCluster,
    KubeRayClusterClientResource,
)
from dagster_ray.kuberay.resources.rayjob import KubeRayInteractiveJob, KubeRayJobClientResource

__all__ = [
    "KubeRayCluster",
    "KubeRayClusterClientResource",
    "RayClusterConfig",
    "KubeRayJobClientResource",
    "RayJobConfig",
    "RayJobSpec",
    "RayClusterSpec",
    "RayClusterConfig",
    "KubeRayInteractiveJob",
]



from __future__ import annotations

import os
from collections.abc import Mapping
from typing import Any, Literal

import dagster as dg
from pydantic import Field

from dagster_ray.kuberay.utils import remove_none_from_dict
from dagster_ray.types import AnyDagsterContext

in_k8s = os.environ.get("KUBERNETES_SERVICE_HOST") is not None


DEFAULT_AUTOSCALER_OPTIONS = {
    "upscalingMode": "Default",
    "idleTimeoutSeconds": 60,
    "env": [],
    "envFrom": [],
    "resources": {
        "limits": {"cpu": "50m", "memory": "0.1Gi"},
        "requests": {"cpu": "50m", "memory": "0.1Gi"},
    },
}
DEFAULT_HEAD_GROUP_SPEC = {
    "serviceType": "ClusterIP",
    "rayStartParams": {},
    "metadata": {
        "labels": {},
        "annotations": {},
    },
    "template": {
        "spec": {
            "imagePullSecrets": [],
            "containers": [
                {
                    "volumeMounts": [
                        {"mountPath": "/tmp/ray", "name": "ray-logs"},
                    ],
                    "name": "head",
                    "imagePullPolicy": "Always",
                },
            ],
            "volumes": [
                {"name": "ray-logs", "emptyDir": {}},
            ],
            "affinity": {},
            "tolerations": [],
            "nodeSelector": {},
        },
    },
}
DEFAULT_WORKER_GROUP_SPECS = [
    {
        "groupName": "workers",
        "replicas": 0,
        "minReplicas": 0,
        "maxReplicas": 1,
        "rayStartParams": {},
        "template": {
            "metadata": {"labels": {}, "annotations": {}},
            "spec": {
                "imagePullSecrets": [],
                "containers": [
                    {
                        "volumeMounts": [{"mountPath": "/tmp/ray", "name": "ray-logs"}],
                        "name": "worker",
                        "imagePullPolicy": "Always",
                    }
                ],
                "volumes": [
                    {"name": "ray-logs", "emptyDir": {}},
                ],
                "affinity": {},
                "tolerations": [],
                "nodeSelector": {},
            },
        },
    }
]


MISSING_IMAGE_MESSAGE = "Image is missing from the `RayCluster` spec, from the top-level Dagster resource config, and the Dagster run does not have a `dagster/image` tag. Please use one of these options to specify the image."


class RayClusterSpec(dg.PermissiveConfig):
    """[RayCluster spec](https://ray-project.github.io/kuberay/reference/api/#rayclusterspec) configuration options. A few sensible defaults are provided for convenience."""

    suspend: bool | None = None
    managed_by: str | None = None
    autoscaler_options: dict[str, Any] = DEFAULT_AUTOSCALER_OPTIONS
    head_service_annotations: dict[str, str] | None = None
    enable_in_tree_autoscaling: bool = False
    gcs_fault_tolerance_options: dict[str, Any] | None = None
    head_group_spec: dict[str, Any] = DEFAULT_HEAD_GROUP_SPEC
    ray_version: str | None = None
    worker_group_specs: list[dict[str, Any]] = DEFAULT_WORKER_GROUP_SPECS

    def to_k8s(
        self,
        context: AnyDagsterContext,
        image: str | None = None,  # is injected into headgroup and workergroups, unless already specified there
        env_vars: Mapping[str, str] | None = None,
    ) -> dict[str, Any]:
        """Convert into Kubernetes manifests in camelCase format and inject additional information"""

        assert context.log is not None

        # TODO: inject self.redis_port and self.dashboard_port into the RayCluster configuration
        # TODO: auto-apply some tags from dagster-k8s/config

        head_group_spec = self.head_group_spec.copy()
        worker_group_specs = self.worker_group_specs.copy()

        k8s_env_vars: list[dict[str, Any]] = []

        if env_vars:
            for key, value in env_vars.items():
                k8s_env_vars.append({"name": key, "value": value})

        def update_group_spec(group_spec: dict[str, Any]):
            # TODO: only inject if the container has a `dagster.io/inject-image` annotation or smth
            if group_spec["template"]["spec"]["containers"][0].get("image") is None:
                if image is None:
                    raise ValueError(MISSING_IMAGE_MESSAGE)
                else:
                    group_spec["template"]["spec"]["containers"][0]["image"] = image

            for container in group_spec["template"]["spec"]["containers"]:
                container["env"] = container.get("env", []) + k8s_env_vars

        update_group_spec(head_group_spec)
        for worker_group_spec in worker_group_specs:
            update_group_spec(worker_group_spec)

        return remove_none_from_dict(
            {
                "enableInTreeAutoscaling": self.enable_in_tree_autoscaling,
                "autoscalerOptions": self.autoscaler_options,
                "headGroupSpec": head_group_spec,
                "workerGroupSpecs": worker_group_specs,
                "suspend": self.suspend,
                "managedBy": self.managed_by,
                "headServiceAnnotations": self.head_service_annotations,
                "gcsFaultToleranceOptions": self.gcs_fault_tolerance_options,
                "rayVersion": self.ray_version,
            }
        )


class RayClusterConfig(dg.Config):
    kind: str = "RayCluster"
    api_version: str = "ray.io/v1"
    metadata: dict[str, Any] = Field(
        default_factory=dict,
        description="Kubernetes metadata, except the name field can be omitted. In this case it will be generated by `dagster-ray`.",
    )
    spec: RayClusterSpec = Field(default_factory=RayClusterSpec)

    @property
    def namespace(self) -> str:
        return self.metadata.get("namespace", "default")

    def to_k8s(
        self,
        context: AnyDagsterContext,
        image: str | None = None,  # is injected into headgroup and workergroups, unless already specified there
        labels: Mapping[str, str] | None = None,
        annotations: Mapping[str, str] | None = None,
        env_vars: Mapping[str, str] | None = None,
    ) -> dict[str, Any]:
        assert context.log is not None
        """Convert into Kubernetes manifests in camelCase format and inject additional information"""

        labels = labels or {}
        annotations = annotations or {}

        return {
            "apiVersion": self.api_version,
            "kind": self.kind,
            "metadata": remove_none_from_dict(
                {
                    "name": self.metadata.get("name"),
                    "labels": {**(self.metadata.get("labels", {}) or {}), **labels},
                    "annotations": {**self.metadata.get("annotations", {}), **annotations},
                }
            ),
            "spec": self.spec.to_k8s(context=context, image=image, env_vars=env_vars),
        }


class RayJobSpec(dg.PermissiveConfig):
    """[RayJob spec](https://ray-project.github.io/kuberay/reference/api/#rayjobspec) configuration options. A few sensible defaults are provided for convenience."""

    active_deadline_seconds: int = 60 * 60 * 24  # 24 hours
    backoff_limit: int = 0
    ray_cluster_spec: RayClusterSpec | None = Field(default_factory=RayClusterSpec)
    submitter_pod_template: dict[str, Any] | None = None
    metadata: dict[str, Any] | None = None
    cluster_selector: dict[str, str] | None = None
    managed_by: str | None = None
    deletion_strategy: dict[str, Any] | None = Field(
        default_factory=lambda: {"onFailure": {"policy": "DeleteCluster"}, "onSuccess": {"policy": "DeleteCluster"}}
    )
    runtime_env_yaml: str | None = None
    job_id: str | None = None
    submission_mode: Literal["K8sJobMode", "HTTPMode", "InteractiveMode"] = "K8sJobMode"
    entrypoint_resources: str | None = None
    entrypoint_num_cpus: float | None = None
    entrypoint_memory: float | None = None
    entrypoint_num_gpus: float | None = None
    ttl_seconds_after_finished: int | None = 5 * 60  # 5 minutes
    shutdown_after_job_finishes: bool = True
    suspend: bool | None = None

    def to_k8s(
        self,
        context: AnyDagsterContext,
        image: str | None = None,  # is injected into headgroup and workergroups, unless already specified there
        env_vars: Mapping[str, str] | None = None,
    ) -> dict[str, Any]:
        """Convert into Kubernetes manifests in camelCase format and inject additional information"""
        return remove_none_from_dict(
            {
                "activeDeadlineSeconds": self.active_deadline_seconds,
                "backoffLimit": self.backoff_limit,
                "submitterPodTemplate": self.submitter_pod_template,
                "metadata": self.metadata,
                "clusterSelector": self.cluster_selector,
                "managedBy": self.managed_by,
                "deletionStrategy": self.deletion_strategy,
                "runtimeEnvYAML": self.runtime_env_yaml,
                "jobId": self.job_id,
                "submissionMode": self.submission_mode,
                "entrypointResources": self.entrypoint_resources,
                "entrypointNumCpus": self.entrypoint_num_cpus,
                "entrypointMemory": self.entrypoint_memory,
                "entrypointNumGpus": self.entrypoint_num_gpus,
                "ttlSecondsAfterFinished": self.ttl_seconds_after_finished,
                "shutdownAfterJobFinishes": self.shutdown_after_job_finishes,
                "suspend": self.suspend,
                "rayClusterSpec": self.ray_cluster_spec.to_k8s(context=context, image=image, env_vars=env_vars)
                if self.ray_cluster_spec is not None
                else None,
            }
        )


class RayJobConfig(dg.Config):
    kind: str = "RayJob"
    api_version: str = "ray.io/v1"
    metadata: dict[str, Any] = Field(
        default_factory=dict,
        description="Kubernetes metadata, except the name field can be omitted. In this case it will be generated by `dagster-ray`.",
    )
    spec: RayJobSpec = Field(default_factory=RayJobSpec)

    @property
    def namespace(self) -> str:
        return self.metadata.get("namespace", "default")

    def to_k8s(
        self,
        context: AnyDagsterContext,
        image: str | None = None,  # is injected into headgroup and workergroups, unless already specified there
        labels: Mapping[str, str] | None = None,
        annotations: Mapping[str, str] | None = None,
        env_vars: Mapping[str, str] | None = None,
    ) -> dict[str, Any]:
        """Convert into Kubernetes manifests in camelCase format and inject additional information"""

        labels = labels or {}
        annotations = annotations or {}

        return {
            "apiVersion": self.api_version,
            "kind": self.kind,
            "metadata": remove_none_from_dict(
                {
                    "name": self.metadata.get("name"),
                    "labels": {**(self.metadata.get("labels", {}) or {}), **labels},
                    "annotations": {**self.metadata.get("annotations", {}), **annotations},
                }
            ),
            "spec": self.spec.to_k8s(
                context=context,
                image=image,
                env_vars=env_vars,
            ),
        }



from __future__ import annotations

import hashlib
import random
import re
import string
from typing import Any

import dagster._check as check

_INVALID_CHARS_PATTERN = re.compile(r"[^a-zA-Z0-9\-_.]")
_LEADING_TRAILING_NON_ALNUM_PATTERN = re.compile(r"^[^a-zA-Z0-9]+|[^a-zA-Z0-9]+$")


def normalize_k8s_label_values(labels: dict[str, str]) -> dict[str, str]:
    """Normalize label values to comply with Kubernetes requirements.

    K8s label values must match: (([A-Za-z0-9][-A-Za-z0-9_.]*)?[A-Za-z0-9])?
    - Start and end with alphanumeric
    - Middle can contain alphanumeric, -, _, .
    - Max 63 characters
    - Empty string is valid

    Additionally, key starting with `dagster/` are replaced with `dagster.io/`
    """
    normalized = {}

    for key, value in labels.items():
        if not value:  # Empty string is valid
            normalized[key] = value
            continue

        # Replace common email symbols for better readability
        value = value.replace("@", "-at-")
        value = value.replace("+", "-plus-")

        # Replace remaining invalid characters with hyphen
        value = _INVALID_CHARS_PATTERN.sub("-", value)

        # Collapse multiple consecutive hyphens into a single hyphen
        value = re.sub(r"-+", "-", value)

        # Remove leading/trailing non-alphanumeric characters
        value = _LEADING_TRAILING_NON_ALNUM_PATTERN.sub("", value)

        # Truncate to 63 characters
        normalized[key] = value[:63]

    # replace keys starting with `dagster.io/` with `dagster/`
    normalized_with_fixed_dagster_keys = {}
    for key, value in normalized.items():
        if key.startswith("dagster.io/"):
            normalized_with_fixed_dagster_keys[f"dagster/{key[11:]}"] = value
        else:
            normalized_with_fixed_dagster_keys[key] = value
    return normalized_with_fixed_dagster_keys


def remove_none_from_dict(d: dict[str, Any]) -> dict[str, Any]:
    return {k: v for k, v in d.items() if v is not None}


def get_k8s_object_name(run_id: str, step_key: str | None = None):
    """Creates a unique (short!) identifier to name k8s objects based on run ID and step key(s).

    K8s Job names are limited to 63 characters, because they are used as labels. For more info, see:

    https://kubernetes.io/docs/concepts/overview/working-with-objects/names/
    """
    check.str_param(run_id, "run_id")
    check.opt_str_param(step_key, "step_key")
    if not step_key:
        letters = string.ascii_lowercase
        step_key = "".join(random.choice(letters) for i in range(20))

    # Creates 32-bit signed int, so could be negative
    name_hash = hashlib.md5((run_id + step_key).encode("utf-8"))

    return name_hash.hexdigest()[:8]



from __future__ import annotations

import contextlib
from abc import ABC, abstractmethod
from collections.abc import Generator
from typing import TYPE_CHECKING, Any, Literal, Optional, cast

import dagster as dg
from pydantic import Field, PrivateAttr

# yes, `python-client` is actually the KubeRay package name
# https://github.com/ray-project/kuberay/issues/2078
from requests.exceptions import ConnectionError
from tenacity import retry, retry_if_exception_type, stop_after_delay
from typing_extensions import Self

from dagster_ray._base.utils import get_dagster_tags
from dagster_ray.config import RayDataExecutionOptions
from dagster_ray.types import AnyDagsterContext
from dagster_ray.utils import _process_dagster_env_vars, get_current_job_id

if TYPE_CHECKING:
    from ray._private.worker import BaseContext as RayBaseContext  # noqa


class Lifecycle(dg.Config):
    create: bool = Field(
        default=True,
        description="Whether to create the resource. If set to `False`, the user can manually call `.create` instead.",
    )
    wait: bool = Field(
        default=True,
        description="Whether to wait for the remote Ray cluster to become ready to accept connections. If set to `False`, the user can manually call `.wait` instead.",
    )
    connect: bool = Field(
        default=True,
        description="Whether to run `ray.init` against the remote Ray cluster. If set to `False`, the user can manually call `.connect` instead.",
    )
    cleanup: Literal["never", "always", "on_exception"] = Field(
        default="always",
        description="Resource cleanup policy. Determines when the resource should be deleted after Dagster step execution or during interruption.",
    )


class BaseRayResource(dg.ConfigurableResource, ABC):
    """Base class for Ray Resources providing a common interface for Ray cluster management.

    This abstract base class defines the interface that all Ray resources must implement,
    providing a backend-agnostic way to interact with Ray clusters. Concrete implementations
    include LocalRay for local development and KubeRay resources for Kubernetes deployments.

    The BaseRayResource handles the lifecycle of Ray clusters including creation, connection,
    and cleanup, with configurable policies for each stage.

    Example:
        Use as a type annotation for backend-agnostic code:
        ```python
        from dagster import asset
        from dagster_ray import RayResource

        @asset
        def my_asset(ray_cluster: RayResource):
            # Works with any Ray backend
            import ray
            return ray.get(ray.put("hello"))
        ```
    Example:
        Manual lifecycle management:
        ```python
        from dagster_ray import Lifecycle

        ray_resource = SomeRayResource(
            lifecycle=Lifecycle(
                create=False,  # Don't auto-create
                connect=False  # Don't auto-connect
            )
        )
        ```

    Note:
        This is an abstract class and cannot be instantiated directly. Use concrete
        implementations like LocalRay or KubeRayCluster instead.
    """

    lifecycle: Lifecycle = Field(default_factory=Lifecycle, description="Actions to perform during resource setup.")
    timeout: float = Field(default=600.0, description="Timeout for Ray readiness in seconds")
    ray_init_options: dict[str, Any] = Field(
        default_factory=dict,
        description="Additional keyword arguments to pass to `ray.init()` call, such as `runtime_env`, `num_cpus`, etc. Dagster's `EnvVar` is supported. More details in [Ray docs](https://docs.ray.io/en/latest/ray-core/api/doc/ray.init.html).",
    )
    data_execution_options: RayDataExecutionOptions = Field(default_factory=RayDataExecutionOptions)
    redis_port: int = Field(
        default=10001, description="Redis port for connection. Make sure to match with the actual available port."
    )
    dashboard_port: int = Field(
        default=8265, description="Dashboard port for connection. Make sure to match with the actual available port."
    )
    env_vars: dict[str, str] | None = Field(
        default_factory=dict, description="Environment variables to pass to the Ray cluster."
    )
    enable_tracing: bool = Field(
        default=False,
        description="Enable tracing: inject `RAY_PROFILING=1` and `RAY_task_events_report_interval_ms=0` into the Ray cluster configuration. This allows using `ray.timeline()` to fetch recorded task events. Learn more: https://docs.ray.io/en/latest/ray-core/api/doc/ray.timeline.html#ray-timeline",
    )
    enable_actor_task_logging: bool = Field(
        default=False,
        description="Enable actor task logging: inject `RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING=1` into the Ray cluster configuration.",
    )
    enable_debug_post_mortem: bool = Field(
        default=False,
        description="Enable post-mortem debugging: inject `RAY_DEBUG_POST_MORTEM=1` into the Ray cluster configuration. Learn more: https://docs.ray.io/en/latest/ray-observability/ray-distributed-debugger.html",
    )
    enable_legacy_debugger: bool = Field(
        default=False,
        description="Enable legacy debugger: inject `RAY_DEBUG=legacy` into the Ray cluster configuration. Learn more: https://docs.ray.io/en/latest/ray-observability/user-guides/debug-apps/ray-debugging.html#using-the-ray-debugger",
    )

    _context: RayBaseContext | None = PrivateAttr()

    @property
    def context(self) -> RayBaseContext:
        assert self._context is not None, "RayClusterResource not initialized"
        return self._context

    @property
    @abstractmethod
    def host(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    @property
    def display_name(self) -> str:
        return self.name

    @property
    def ray_address(self) -> str:
        return f"ray://{self.host}:{self.redis_port}"

    @property
    def dashboard_url(self) -> str:
        return f"http://{self.host}:{self.dashboard_port}"

    @property
    def runtime_job_id(self) -> str:
        """
        Returns the Ray Job ID for the current job which was created with `ray.init()`.
        :return:
        """
        return get_current_job_id()

    @contextlib.contextmanager
    def yield_for_execution(self, context: dg.InitResourceContext) -> Generator[Self, None, None]:
        exception_occurred = None
        try:
            if self.lifecycle.create:
                self._create(context)
                if self.lifecycle.wait:
                    self._wait(context)
                    if self.lifecycle.connect:
                        self._connect(context)
            yield self
        except BaseException as e:
            exception_occurred = e
            raise
        finally:
            self.cleanup(context, exception_occurred)

    def _create(self, context: AnyDagsterContext):
        assert context.log is not None
        if not self.created:
            try:
                self.create(context)
                context.log.info(f"Created {self.display_name}.")
            except BaseException:
                context.log.exception(f"Failed to create {self.display_name}")
                raise

    def _wait(self, context: AnyDagsterContext):
        assert context.log is not None
        self._create(context)
        if not self.ready:
            context.log.info(f"Waiting for {self.display_name} to become ready (timeout={self.timeout:.0f}s)...")
            try:
                self.wait(context)
            except BaseException:
                context.log.exception(f"Failed to wait for {self.display_name} readiness")
                raise

    def _connect(self, context: AnyDagsterContext):
        assert context.log is not None
        self._wait(context)
        if not self.connected:
            try:
                self.connect(context)
            except BaseException:
                context.log.exception(f"Failed to connect to {self.display_name}")
                raise
            context.log.info(f"Initialized Ray Client with {self.display_name}")

    def create(self, context: AnyDagsterContext):
        pass

    def wait(self, context: AnyDagsterContext):
        pass

    @retry(stop=stop_after_delay(120), retry=retry_if_exception_type(ConnectionError), reraise=True)
    def connect(self, context: AnyDagsterContext) -> RayBaseContext:
        assert context.log is not None

        import ray

        init_options = _process_dagster_env_vars(self.ray_init_options.copy())

        # cleanup None values from runtime_env.env_vars since Ray doesn't like them

        if "runtime_env" in init_options and "env_vars" in init_options["runtime_env"]:
            init_options["runtime_env"]["env_vars"] = {
                k: v for k, v in init_options["runtime_env"]["env_vars"].items() if v is not None
            }

        init_options["runtime_env"] = init_options.get("runtime_env", {})
        init_options["runtime_env"]["env_vars"] = init_options["runtime_env"].get("env_vars", {})

        for var, value in self.get_env_vars_to_inject().items():
            init_options["runtime_env"]["env_vars"][var] = value

        self.data_execution_options.apply()

        self._context = ray.init(
            address=self.ray_address,
            **init_options,
        )
        self.data_execution_options.apply()
        self.data_execution_options.apply_remote()
        context.log.info("Initialized Ray in client mode!")
        return cast("RayBaseContext", self._context)

    def delete(self, context: AnyDagsterContext):
        pass

    def cleanup(self, context: AnyDagsterContext, exception: Optional[BaseException]):  # noqa: UP007
        assert context.log is not None

        if self.lifecycle.cleanup == "never":
            to_delete = False
        elif not self.created:
            to_delete = False
        elif self.lifecycle.cleanup == "always":
            to_delete = True
        elif self.lifecycle.cleanup == "on_exception":
            to_delete = exception is not None
        else:
            to_delete = False

        if to_delete:
            self.delete(context)
            context.log.info(f'Deleted {self.display_name} according to cleanup policy "{self.lifecycle.cleanup}"')

        if self.connected and hasattr(self, "_context") and self._context is not None:
            self._context.disconnect()

    def get_dagster_tags(self, context: AnyDagsterContext) -> dict[str, str]:
        tags = get_dagster_tags(context)
        return tags

    def get_env_vars_to_inject(self) -> dict[str, str]:
        vars: dict[str, str] = self.env_vars or {}
        if self.enable_debug_post_mortem:
            vars["RAY_DEBUG_POST_MORTEM"] = "1"
        if self.enable_tracing:
            vars["RAY_PROFILING"] = "1"
            vars["RAY_task_events_report_interval_ms"] = "0"
        if self.enable_actor_task_logging:
            vars["RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING"] = "1"
        if self.enable_legacy_debugger:
            vars["RAY_DEBUG"] = "legacy"
        return vars

    @property
    def created(self) -> bool:
        return hasattr(self, "_name") and self._name is not None

    @property
    def ready(self) -> bool:
        return hasattr(self, "_host") and self._host is not None

    @property
    def connected(self) -> bool:
        return hasattr(self, "_context") and self._context is not None



from __future__ import annotations

import re
from abc import abstractmethod
from typing import TYPE_CHECKING
from uuid import uuid4

import dagster as dg
from pydantic import Field, PrivateAttr
from ray._private.worker import BaseContext as RayBaseContext  # noqa

from dagster_ray._base.constants import DEFAULT_DEPLOYMENT_NAME
from dagster_ray.kuberay.utils import get_k8s_object_name
from dagster_ray.types import AnyDagsterContext

if TYPE_CHECKING:
    pass


class BaseKubeRayResourceConfig(dg.Config):
    image: str | None = Field(
        default=None,
        description="Image to inject into the `RayCluster` spec. Defaults to `dagster/image` run tag. Images already provided in the `RayCluster` spec won't be overridden.",
    )
    deployment_name: str = Field(
        default=DEFAULT_DEPLOYMENT_NAME,
        description="Dagster deployment name. Is used as a prefix for the Kubernetes resource name. Dagster Cloud variables are used to determine the default value.",
    )
    poll_interval: float = Field(default=1.0, description="Poll interval for various API requests")

    _host: str = PrivateAttr()

    @property
    @abstractmethod
    def namespace(self) -> str:
        raise NotImplementedError

    def _get_step_name(self, context: AnyDagsterContext) -> str:
        assert isinstance(context.run_id, str)
        assert context.dagster_run is not None

        # try to make the name as short as possible
        cluster_name_prefix = f"dg-{self.deployment_name.replace('-', '')[:8]}-{context.run_id[:8]}"

        dagster_user_email = context.dagster_run.tags.get("user")
        if dagster_user_email is not None:
            cluster_name_prefix += f"-{dagster_user_email.replace('.', '').replace('-', '').split('@')[0][:6]}"

        step_key = str(uuid4())

        name_key = get_k8s_object_name(
            context.run_id,
            step_key,
        )

        step_name = f"{cluster_name_prefix}-{name_key}".lower()
        step_name = re.sub(r"[^-0-9a-z]", "-", step_name)

        return step_name



import sys

# yes, `python-client` is actually the KubeRay package name
# https://github.com/ray-project/kuberay/issues/2078


if sys.version_info >= (3, 11):
    pass
else:
    pass

from ray._private.worker import BaseContext as RayBaseContext  # noqa

from dagster_ray._base.resources import BaseRayResource


class LocalRay(BaseRayResource):
    """
    Dummy Resource.
    Is useful for testing and local development.
    Provides the same interface as actual Resources.
    """

    @property
    def host(self) -> str:
        return "127.0.0.1"

    # the address is overwritten by None since ray.init does not behave as expected
    # when the host is set to localhost or 127.0.0.1
    @property
    def ray_address(self) -> None:  # type: ignore
        return None

    @property
    def name(self) -> str:
        return "LocalRay"



from typing import TYPE_CHECKING, Literal, Optional

import dagster as dg
from pydantic import Field, PrivateAttr
from typing_extensions import override

from dagster_ray._base.resources import BaseRayResource, Lifecycle
from dagster_ray.kuberay.client import RayJobClient
from dagster_ray.kuberay.client.base import load_kubeconfig
from dagster_ray.kuberay.configs import RayJobConfig, RayJobSpec
from dagster_ray.kuberay.resources.base import BaseKubeRayResourceConfig
from dagster_ray.kuberay.utils import normalize_k8s_label_values
from dagster_ray.types import AnyDagsterContext

if TYPE_CHECKING:
    from ray._private.worker import BaseContext as RayBaseContext  # noqa

if TYPE_CHECKING:
    pass


class KubeRayJobClientResource(dg.ConfigurableResource[RayJobClient]):
    """This configurable resource provides a `dagster_ray.kuberay.client.RayJobClient`."""

    kube_context: Optional[str] = None
    kube_config: Optional[str] = None

    def create_resource(self, context: dg.InitResourceContext):
        load_kubeconfig(context=self.kube_context, config_file=self.kube_config)
        return RayJobClient(kube_context=self.kube_context, kube_config=self.kube_config)


class InteractiveRayJobSpec(RayJobSpec):
    """Same as `dagster_ray.kuberay.resources.rayjob.RayJobSpec`, but submission mode has to be `InteractiveMode`"""

    submission_mode: Literal["InteractiveMode"] = "InteractiveMode"  # pyright: ignore[reportIncompatibleVariableOverride]


class InteractiveRayJobConfig(RayJobConfig):
    """Same as `dagster_ray.kuberay.resources.rayjob.RayJobConfig`, but `spec.submission_mode` mode has to be `InteractiveMode`"""

    spec: InteractiveRayJobSpec = Field(default_factory=InteractiveRayJobSpec)  # pyright: ignore[reportIncompatibleVariableOverride]


class KubeRayInteractiveJob(BaseRayResource, BaseKubeRayResourceConfig):
    """
    Provides a `RayJob` for Dagster steps.

    Is the recommended way to run Ray workloads with automatic cluster management. It creates a `RayJob`, connects to it in client mode and sets the `jobId` field. Cleanup is handled by the KubeRay controller or by the resource lifecycle logic.

    Info:
        Image defaults to `dagster/image` run tag.

    Tip:
        Make sure `ray[full]` is available in the image.
    """

    lifecycle: Lifecycle = Field(
        default_factory=lambda: Lifecycle(
            cleanup="on_exception"  # RayJob has it's own lifecycle management so it makes sense to only interfere when the step has been cancelled, otherwise it will be left in Waiting state (see https://github.com/ray-project/kuberay/issues/4037)
        ),
        description="Actions to perform during resource setup.",
    )

    ray_job: InteractiveRayJobConfig = Field(
        default_factory=InteractiveRayJobConfig, description="Configuration for the Kubernetes `RayJob` CR"
    )
    client: dg.ResourceDependency[RayJobClient] = Field(  # pyright: ignore[reportAssignmentType]
        default_factory=KubeRayJobClientResource, description="Kubernetes `RayJob` client"
    )

    log_cluster_conditions: bool = Field(
        default=True,
        description="Whether to log `RayCluster` conditions while waiting for the RayCluster to become ready. For more information, see https://docs.ray.io/en/latest/cluster/kubernetes/user-guides/observability.html#raycluster-status-conditions.",
    )

    _name: str = PrivateAttr()
    _cluster_name: str = PrivateAttr()
    _host: str = PrivateAttr()

    @property
    @override
    def host(self) -> str:
        if not hasattr(self, "_host"):
            raise ValueError(f"{self.__class__.__name__} not initialized")
        return self._host

    @property
    def name(self) -> str:
        if not hasattr(self, "_name"):
            raise ValueError(f"{self.__class__.__name__} is not initialized")
        elif (name := self.ray_job.metadata.get("name")) is not None:
            return name
        else:
            return self._name

    @property
    @override
    def namespace(self) -> str:
        return self.ray_job.namespace

    @property
    @override
    def display_name(self) -> str:
        return f"RayJob {self.namespace}/{self.name}" if self.created else f"RayJob in namespace {self.namespace}"

    @property
    def cluster_name(self) -> str:
        if not hasattr(self, "_cluster_name"):
            raise ValueError(f"{self.__class__.__name__} not initialized")
        else:
            return self._cluster_name

    @override
    def get_dagster_tags(self, context: AnyDagsterContext) -> dict[str, str]:
        tags = super().get_dagster_tags(context=context)
        tags.update({"dagster/deployment": self.deployment_name})
        return tags

    @override
    def create(self, context: AnyDagsterContext):
        assert context.log is not None
        assert context.dagster_run is not None

        self._name = self.ray_job.metadata.get("name") or self._get_step_name(context)

        k8s_manifest = self.ray_job.to_k8s(
            context,
            image=(self.image or context.dagster_run.tags.get("dagster/image")),
            labels=normalize_k8s_label_values(self.get_dagster_tags(context)),
        )

        k8s_manifest["metadata"]["name"] = self.name

        if not self.client.get(
            name=self.name,
            namespace=self.namespace,
        ):
            resource = self.client.create(body=k8s_manifest, namespace=self.namespace)

            if not resource:
                raise RuntimeError(f"Couldn't create {self.display_name}")

    @override
    def wait(self, context: AnyDagsterContext):
        assert context.log is not None
        assert context.dagster_run is not None

        self.client.wait_until_ready(
            name=self.name,
            namespace=self.namespace,
            log_cluster_conditions=self.log_cluster_conditions,
            timeout=self.timeout,
            poll_interval=self.poll_interval,
        )
        self._cluster_name = self.client.get_ray_cluster_name(
            self.name, self.namespace, timeout=self.timeout, poll_interval=self.poll_interval
        )
        self._host = self.client.ray_cluster_client.get_status(
            name=self.cluster_name, namespace=self.namespace, timeout=self.timeout, poll_interval=self.poll_interval
        )[  # pyright: ignore
            "head"
        ]["serviceIP"]

        msg = f"RayJob {self.namespace}/{self.name} has created a RayCluster {self.namespace}/{self.cluster_name}! Connection command:\n"
        msg += (
            f"kubectl -n {self.namespace} port-forward svc/{self.cluster_name}-head-svc 8265:8265 6379:6379 10001:10001"
        )

        context.log.info(msg)

    @override
    def connect(self, context: AnyDagsterContext) -> "RayBaseContext":
        """Connect to Ray and set `RayJobSpec.jobId` to bind the client session to the `RayJob` CR. Requires KubeRay 1.3.0.

        This procedure is described in https://github.com/ray-project/kuberay/pull/2364
        """
        ray_context = super().connect(context)

        # now point the RayJob at our ray job
        self.client.update(
            name=self.name,
            namespace=self.namespace,
            body={"spec": {"jobId": self.runtime_job_id}},
        )

        return ray_context

    @override
    def delete(self, context: AnyDagsterContext):
        self.client.delete(self.name, namespace=self.namespace)



from typing import Union

import dagster as dg
from typing_extensions import TypeAlias

OpOrAssetExecutionContext: TypeAlias = Union[dg.OpExecutionContext, dg.AssetExecutionContext]

AnyDagsterContext: TypeAlias = Union[dg.OpExecutionContext, dg.AssetExecutionContext, dg.InitResourceContext]



from __future__ import annotations

import dagster as dg

from dagster_ray._base.constants import DEFAULT_DEPLOYMENT_NAME


def get_dagster_tags(
    context: dg.InitResourceContext | dg.OpExecutionContext | dg.AssetExecutionContext,
    extra_tags: dict[str, str] | None = None,
) -> dict[str, str]:
    """
    Returns a dictionary with common Dagster tags.
    """
    assert context.dagster_run is not None

    labels: dict[str, str] = {
        "dagster/deployment": DEFAULT_DEPLOYMENT_NAME,  # TODO: this should come from extra_tags,
        **(extra_tags or {}),
    }

    if isinstance(context, dg.InitResourceContext):
        labels.update(
            **context.dagster_run.dagster_execution_info,
        )

        for resource_key, resource_def in context.all_resource_defs.items():
            if resource_def is context.resource_def:
                # inject the resource key used for this KubeRay resource
                # this enables e.g reusing the same `RayCluster` across Dagster steps that require it without recreating the cluster
                labels["dagster/resource-key"] = resource_key
    else:
        labels.update(
            **context.run.dagster_execution_info,
        )

    return labels
